{% extends "page.html" %}

{% block title %}
    {{ _('Search') }}
{% endblock %}

{% block breadcrumb_content %}
    <li class="active">
        <a class="active" href="/search">Search</a>
    </li>
{% endblock %}

{% block pre_primary %}
    {% asset 'ckanext-versioned-datastore/search' %}

    <article class="module">
        <div class="module-content">
            {% block search_content %}
                <div id="searchApp" data-module="vds_search" class="hide-until-loaded">
                    <div class="module module-search module-narrow module-shallow">
                        <form class="module-content search-form" id="searchForm">
                            <div class="search-input control-group search-giant">
                                <label for="all" class="sr-only">Search</label>
                                <input type="text" class="search" name="all" id="all"
                                        value="" autocomplete="off"
                                        placeholder="Search all fields"/>
                                <button type="submit">
                                    <i class="fas fa-search"></i>
                                    <span class="sr-only">{{ _('Search') }}</span>
                                </button>
                            </div>
                        </form>
                    </div>
                    <div id="result"></div>
                </div>
            {% endblock %}
        </div>
    </article>
{% endblock %}

{% block secondary %}
{% endblock %}
