adjectives = [
    'acceptable',
    'according',
    'accurate',
    'active',
    'actual',
    'additional',
    'administrative',
    'adult',
    'afraid',
    'afternoon',
    'aggressive',
    'alone',
    'alternative',
    'amazing',
    'angry',
    'annual',
    'anxious',
    'any',
    'appropriate',
    'asleep',
    'automatic',
    'available',
    'aware',
    'background',
    'basic',
    'beautiful',
    'best',
    'big',
    'bitter',
    'boring',
    'brave',
    'brief',
    'bright',
    'brilliant',
    'broad',
    'budget',
    'business',
    'busy',
    'calm',
    'capable',
    'careful',
    'certain',
    'chance',
    'cheap',
    'chemical',
    'civil',
    'classic',
    'clean',
    'clear',
    'cold',
    'comfortable',
    'commercial',
    'common',
    'competitive',
    'complete',
    'complex',
    'comprehensive',
    'confident',
    'conscious',
    'consistent',
    'constant',
    'content',
    'cool',
    'correct',
    'creative',
    'critical',
    'cultural',
    'curious',
    'current',
    'cute',
    'dangerous',
    'dark',
    'dear',
    'decent',
    'deep',
    'dependent',
    'designer',
    'desperate',
    'different',
    'difficult',
    'direct',
    'dirty',
    'distinct',
    'double',
    'downtown',
    'dramatic',
    'drunk',
    'dry',
    'easy',
    'educational',
    'effective',
    'efficient',
    'electrical',
    'electronic',
    'embarrassed',
    'emergency',
    'emotional',
    'empty',
    'enough',
    'entire',
    'environmental',
    'equal',
    'equivalent',
    'every',
    'exact',
    'excellent',
    'exciting',
    'expensive',
    'expert',
    'express',
    'external',
    'extra',
    'extreme',
    'fair',
    'false',
    'familiar',
    'famous',
    'fast',
    'fat',
    'federal',
    'final',
    'financial',
    'fine',
    'firm',
    'first',
    'fit',
    'flat',
    'formal',
    'former',
    'forward',
    'free',
    'frequent',
    'fresh',
    'friendly',
    'full',
    'fun',
    'funny',
    'future',
    'general',
    'glad',
    'glass',
    'global',
    'gold',
    'good',
    'grand',
    'great',
    'green',
    'gross',
    'guilty',
    'happy',
    'healthy',
    'heavy',
    'helpful',
    'high',
    'historical',
    'honest',
    'huge',
    'hungry',
    'ideal',
    'ill',
    'illegal',
    'immediate',
    'important',
    'impossible',
    'impressive',
    'independent',
    'individual',
    'inevitable',
    'informal',
    'initial',
    'inner',
    'intelligent',
    'interesting',
    'internal',
    'international',
    'junior',
    'key',
    'kind',
    'kitchen',
    'known',
    'large',
    'last',
    'late',
    'latter',
    'leading',
    'least',
    'left',
    'legal',
    'little',
    'local',
    'logical',
    'lonely',
    'long',
    'lost',
    'loud',
    'low',
    'lower',
    'lucky',
    'main',
    'major',
    'many',
    'massive',
    'material',
    'maximum',
    'mean',
    'medical',
    'medium',
    'middle',
    'minor',
    'minute',
    'mobile',
    'mother',
    'narrow',
    'nasty',
    'national',
    'natural',
    'nearby',
    'neat',
    'necessary',
    'negative',
    'neither',
    'nervous',
    'new',
    'nice',
    'normal',
    'novel',
    'numerous',
    'objective',
    'obvious',
    'odd',
    'official',
    'ok',
    'old',
    'open',
    'opening',
    'opposite',
    'orange',
    'ordinary',
    'original',
    'other',
    'overall',
    'parking',
    'particular',
    'party',
    'past',
    'patient',
    'perfect',
    'personal',
    'physical',
    'plastic',
    'pleasant',
    'political',
    'poor',
    'popular',
    'positive',
    'possible',
    'potential',
    'powerful',
    'practical',
    'pregnant',
    'present',
    'pretend',
    'pretty',
    'previous',
    'primary',
    'prior',
    'private',
    'prize',
    'professional',
    'proper',
    'proud',
    'psychological',
    'public',
    'pure',
    'purple',
    'quick',
    'quiet',
    'rare',
    'raw',
    'ready',
    'real',
    'realistic',
    'reasonable',
    'recent',
    'regular',
    'relevant',
    'remarkable',
    'remote',
    'representative',
    'resident',
    'responsible',
    'rich',
    'rough',
    'round',
    'routine',
    'royal',
    'sad',
    'safe',
    'salty',
    'scared',
    'sea',
    'secret',
    'secure',
    'select',
    'senior',
    'sensitive',
    'serious',
    'several',
    'severe',
    'sharp',
    'short',
    'sick',
    'significant',
    'silly',
    'silver',
    'similar',
    'simple',
    'single',
    'slight',
    'slow',
    'small',
    'smart',
    'smooth',
    'soft',
    'solid',
    'sorry',
    'southern',
    'spare',
    'special',
    'specialist',
    'specific',
    'spiritual',
    'square',
    'standard',
    'stock',
    'strange',
    'strict',
    'strong',
    'substantial',
    'successful',
    'such',
    'sudden',
    'sufficient',
    'suitable',
    'super',
    'suspicious',
    'sweet',
    'swimming',
    'tall',
    'technical',
    'temporary',
    'terrible',
    'thin',
    'tiny',
    'top',
    'total',
    'tough',
    'traditional',
    'training',
    'trick',
    'true',
    'typical',
    'ugly',
    'unable',
    'unfair',
    'unhappy',
    'unique',
    'united',
    'unlikely',
    'unusual',
    'upper',
    'upset',
    'used',
    'useful',
    'usual',
    'valuable',
    'various',
    'vast',
    'visible',
    'visual',
    'warm',
    'waste',
    'weak',
    'weekly',
    'weird',
    'whole',
    'wide',
    'wild',
    'willing',
    'winter',
    'wise',
    'wonderful',
    'wooden',
    'work',
    'working',
    'worth',
    'wrong',
    'young',
]

animals = [
    'aardvark',
    'albatross',
    'alligator',
    'angelfish',
    'ant',
    'anteater',
    'antelope',
    'armadillo',
    'axolotl',
    'baboon',
    'badger',
    'bandicoot',
    'barnacle',
    'bat',
    'bear',
    'beaver',
    'beetle',
    'bird',
    'bison',
    'bobcat',
    'bonobo',
    'budgerigar',
    'buffalo',
    'butterfly',
    'camel',
    'capybara',
    'cassowary',
    'cat',
    'caterpillar',
    'catfish',
    'centipede',
    'chameleon',
    'cheetah',
    'chicken',
    'chimpanzee',
    'chinchilla',
    'chipmunk',
    'coral',
    'cougar',
    'cow',
    'coyote',
    'crab',
    'crane',
    'crocodile',
    'cuttlefish',
    'deer',
    'dingo',
    'dodo',
    'dog',
    'dolphin',
    'donkey',
    'dormouse',
    'dragonfly',
    'duck',
    'dugong',
    'eagle',
    'earwig',
    'elephant',
    'emu',
    'falcon',
    'ferret',
    'fish',
    'flamingo',
    'fly',
    'fox',
    'frog',
    'gecko',
    'gerbil',
    'gibbon',
    'giraffe',
    'goat',
    'goose',
    'gorilla',
    'grasshopper',
    'guppy',
    'hamster',
    'hare',
    'hedgehog',
    'heron',
    'hippopotamus',
    'horse',
    'hummingbird',
    'hyena',
    'iguana',
    'impala',
    'jackal',
    'jaguar',
    'jellyfish',
    'kangaroo',
    'kingfisher',
    'kiwi',
    'koala',
    'ladybird',
    'lemming',
    'lemur',
    'leopard',
    'lion',
    'lizard',
    'llama',
    'lobster',
    'lynx',
    'macaw',
    'magpie',
    'manatee',
    'mayfly',
    'meerkat',
    'millipede',
    'mole',
    'mongoose',
    'monkey',
    'moorhen',
    'moose',
    'moth',
    'mouse',
    'mule',
    'newt',
    'nightingale',
    'ocelot',
    'octopus',
    'orangutan',
    'ostrich',
    'otter',
    'oyster',
    'pangolin',
    'panther',
    'parrot',
    'peacock',
    'pelican',
    'penguin',
    'pheasant',
    'pig',
    'piranha',
    'platypus',
    'porcupine',
    'possum',
    'prawn',
    'puffin',
    'puma',
    'quail',
    'rabbit',
    'raccoon',
    'rat',
    'rattlesnake',
    'reindeer',
    'rhinoceros',
    'robin',
    'salamander',
    'scorpion',
    'seahorse',
    'seal',
    'sheep',
    'shrimp',
    'skunk',
    'sloth',
    'snail',
    'snake',
    'sparrow',
    'sponge',
    'squid',
    'squirrel',
    'starfish',
    'stingray',
    'swan',
    'tapir',
    'termite',
    'tiger',
    'tortoise',
    'toucan',
    'turkey',
    'vulture',
    'wallaby',
    'walrus',
    'warthog',
    'wasp',
    'weasel',
    'wildebeest',
    'wolf',
    'wolverine',
    'wombat',
    'woodlouse',
    'woodpecker',
    'yak',
    'zebra',
]

adverbs = [
    'absentmindedly',
    'acceptably',
    'accidentally',
    'accordingly',
    'accurately',
    'actively',
    'actually',
    'additionally',
    'administratively',
    'aggressively',
    'alternatively',
    'always',
    'amazingly',
    'angrily',
    'annually',
    'anxiously',
    'appropriately',
    'automatically',
    'awkwardly',
    'badly',
    'basically',
    'beautifully',
    'bitterly',
    'blindly',
    'boastfully',
    'boldly',
    'bravely',
    'briefly',
    'brightly',
    'brilliantly',
    'broadly',
    'busily',
    'calmly',
    'capably',
    'carefully',
    'certainly',
    'cheaply',
    'cheerfully',
    'chemically',
    'classically',
    'cleanly',
    'clearly',
    'coldly',
    'comfortably',
    'commercially',
    'commonly',
    'competitively',
    'completely',
    'comprehensively',
    'confidently',
    'consciously',
    'consistently',
    'constantly',
    'continually',
    'coolly',
    'correctly',
    'coyly',
    'creatively',
    'critically',
    'culturally',
    'curiously',
    'currently',
    'dangerously',
    'darkly',
    'dearly',
    'decently',
    'deeply',
    'defiantly',
    'deftly',
    'deliberately',
    'delightfully',
    'desperately',
    'devotedly',
    'differently',
    'difficultly',
    'directly',
    'dirtily',
    'distinctly',
    'doubly',
    'doubtfully',
    'dramatically',
    'dryly',
    'dutifully',
    'eagerly',
    'easily',
    'educationally',
    'effectively',
    'efficiently',
    'electrically',
    'electronically',
    'elegantly',
    'emotionally',
    'enormously',
    'entirely',
    'environmentally',
    'equally',
    'essentially',
    'evenly',
    'eventually',
    'exactly',
    'excellently',
    'excitingly',
    'expensively',
    'expertly',
    'expressly',
    'externally',
    'extremely',
    'fairly',
    'faithfully',
    'falsely',
    'famously',
    'federally',
    'ferociously',
    'fiercely',
    'finally',
    'financially',
    'finely',
    'firmly',
    'firstly',
    'flatly',
    'foolishly',
    'formally',
    'formerly',
    'fortunately',
    'freely',
    'frequently',
    'freshly',
    'friendlily',
    'fully',
    'funnily',
    'generally',
    'generously',
    'gladly',
    'gleefully',
    'globally',
    'goodly',
    'gracefully',
    'grandly',
    'greatly',
    'grossly',
    'happily',
    'hastily',
    'healthily',
    'heavily',
    'helpfully',
    'highly',
    'historically',
    'honestly',
    'hopefully',
    'hopelessly',
    'hugely',
    'hungrily',
    'ideally',
    'illegally',
    'immediately',
    'importantly',
    'impossibly',
    'impressively',
    'independently',
    'individually',
    'inevitably',
    'informally',
    'initially',
    'innerly',
    'innocently',
    'inquisitively',
    'intelligently',
    'intensely',
    'interestingly',
    'internally',
    'internationally',
    'irritably',
    'jealously',
    'joyfully',
    'justly',
    'kindly',
    'knowingly',
    'largely',
    'lastly',
    'lately',
    'latterly',
    'lazily',
    'legally',
    'locally',
    'logically',
    'longingly',
    'loosely',
    'loudly',
    'lowly',
    'loyally',
    'luckily',
    'mainly',
    'majorly',
    'massively',
    'materially',
    'medically',
    'merrily',
    'minutely',
    'mockingly',
    'mortally',
    'motherly',
    'mysteriously',
    'narrowly',
    'nastily',
    'nationally',
    'naturally',
    'neatly',
    'necessarily',
    'negatively',
    'nervously',
    'never',
    'newly',
    'nicely',
    'noisily',
    'normally',
    'obediently',
    'objectively',
    'obnoxiously',
    'obviously',
    'occasionally',
    'oddly',
    'offensively',
    'officially',
    'often',
    'only',
    'openly',
    'ordinarily',
    'originally',
    'partially',
    'particularly',
    'patiently',
    'perfectly',
    'personally',
    'physically',
    'plastically',
    'pleasantly',
    'politely',
    'politically',
    'poorly',
    'popularly',
    'positively',
    'possibly',
    'potentially',
    'powerfully',
    'practically',
    'presently',
    'prettily',
    'previously',
    'primarily',
    'privately',
    'professionally',
    'promptly',
    'properly',
    'proudly',
    'psychologically',
    'publicly',
    'punctually',
    'purely',
    'purply',
    'quickly',
    'quietly',
    'quizzically',
    'rapidly',
    'rarely',
    'readily',
    'realistically',
    'really',
    'reasonably',
    'recently',
    'recklessly',
    'regrettably',
    'regularly',
    'remarkably',
    'remotely',
    'repeatedly',
    'responsibly',
    'richly',
    'roughly',
    'roundly',
    'routinely',
    'royally',
    'rudely',
    'sadly',
    'safely',
    'saltily',
    'scarcely',
    'scarily',
    'secretly',
    'securely',
    'seldom',
    'selfishly',
    'sensitively',
    'seriously',
    'severally',
    'severely',
    'shakily',
    'sharply',
    'shortly',
    'sickly',
    'significantly',
    'silently',
    'similarly',
    'simply',
    'singly',
    'slightly',
    'slowly',
    'smartly',
    'smoothly',
    'softly',
    'solemnly',
    'solidly',
    'sometimes',
    'specially',
    'specifically',
    'speedily',
    'spiritually',
    'squarely',
    'sternly',
    'strangely',
    'strictly',
    'strongly',
    'substantially',
    'successfully',
    'suddenly',
    'sufficiently',
    'suitably',
    'suspiciously',
    'sweetly',
    'swimmingly',
    'sympathetically',
    'tally',
    'technically',
    'tediously',
    'temporarily',
    'terribly',
    'thankfully',
    'thinly',
    'totally',
    'traditionally',
    'tremendously',
    'triumphantly',
    'typically',
    'unbearably',
    'unexpectedly',
    'unfairly',
    'unhappily',
    'uniquely',
    'unusually',
    'usefully',
    'usually',
    'vaguely',
    'valiantly',
    'variously',
    'vastly',
    'victoriously',
    'violently',
    'virtually',
    'visibly',
    'visually',
    'vivaciously',
    'warmly',
    'weakly',
    'wearily',
    'weirdly',
    'wholly',
    'widely',
    'wildly',
    'willingly',
    'wisely',
    'woefully',
    'wonderfully',
    'woodenly',
    'wrongly',
]

plants = [
    'almond',
    'apricot',
    'bamboo',
    'banana',
    'baobab',
    'bean',
    'beech',
    'birch',
    'bluebell',
    'cabbage',
    'carrot',
    'cedar',
    'cherry',
    'chestnut',
    'clover',
    'coconut',
    'coffee',
    'elderberry',
    'eucalyptus',
    'fennel',
    'fern',
    'fig',
    'flax',
    'foxglove',
    'fuchsia',
    'garlic',
    'grapefruit',
    'grass',
    'hydrangea',
    'juniper',
    'knotweed',
    'lavender',
    'leek',
    'lemon',
    'lettuce',
    'lupin',
    'magnolia',
    'mahogany',
    'maize',
    'mango',
    'maple',
    'mesquite',
    'milkweed',
    'mulberry',
    'nettle',
    'nightshade',
    'oak',
    'onion',
    'orange',
    'orchid',
    'parsley',
    'parsnip',
    'pea',
    'peach',
    'peanut',
    'pine',
    'pineapple',
    'pistachio',
    'plantain',
    'poplar',
    'potato',
    'primrose',
    'ragweed',
    'raspberry',
    'rhubarb',
    'rice',
    'rocket',
    'rye',
    'saffron',
    'snowdrop',
    'strawberry',
    'sugarcane',
    'sunflower',
    'sugarplum',
    'sycamore',
    'tea',
    'thistle',
    'thyme',
    'tomato',
    'tulip',
    'violet',
    'walnut',
    'wheat',
    'yarrow',
    'yam',
]

list_one = adjectives + adverbs
list_two = adjectives
list_three = animals + plants
