records = [
    {
        'scientificName': 'Boops boops',
        'img': 'img-url-here',
        'urlSlug': 'boops-boops',
        'emptyField': 'not actually empty',
        'group': 'a',
    },
    {
        'scientificName': 'Felis catus',
        'img': 'cat-picture-here',
        'urlSlug': 'felis-catus',
        'group': 'b',
    },
]

records_addtl = [
    {
        'scientificName': 'Gigantiops destructor',
        'img': 'add-picture-here',
        'urlSlug': 'g-destructor',
        'colour': 'green',
        'group': 'b',
    },
    {
        'scientificName': 'Tyrannosaurus rex',
        'img': 'img-url-here',
        'urlSlug': 't-rex',
        'colour': 'blue',
        'group': 'b',
    },
]
