{% ckan_extends %}

{{ super() }}

{% block content_primary_nav %}
  {{ h.build_nav_icon('resource.edit', _('Edit resource'), id=pkg.name, resource_id=res.id) }}
  {{ h.build_nav_icon('datastore.resource_data', _('DataStore'), package_name=pkg.name, resource_id=res.id) }}
  {{ h.build_nav_icon('resource.views', _('Views'), id=pkg.name, resource_id=res.id) }}
{% endblock %}
